package net.knarcraft.dynmapcitizens.trait;

import me.blackvein.quests.QuestsAPI;
import me.blackvein.quests.quests.IQuest;
import me.blackvein.quests.quests.IStage;
import net.knarcraft.dynmapcitizens.DynmapCitizens;
import net.knarcraft.dynmapcitizens.Icon;
import net.knarcraft.dynmapcitizens.UpdateRate;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.World;
import org.dynmap.DynmapAPI;
import org.dynmap.markers.CircleMarker;
import org.dynmap.markers.GenericMarker;
import org.dynmap.markers.MarkerIcon;
import org.dynmap.markers.MarkerSet;

import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;

/**
 * A handler class for the quests trait
 */
public class QuestsHandler extends AbstractTraitHandler {

    private QuestsAPI questsAPI;
    private MarkerSet questMarkerSet;
    private MarkerSet questAreaMarkerSet;
    private Map<Icon, MarkerIcon> markerIcons;

    @Override
    public void initialize() {
        questsAPI = (QuestsAPI) Bukkit.getServer().getPluginManager().getPlugin("Quests");
        DynmapAPI dynmapAPI = DynmapCitizens.getInstance().getDynmapAPI();
        markerIcons = DynmapCitizens.getInstance().getMarkerIcons();
        if (questsAPI != null) {
            questMarkerSet = getMarkerSet(dynmapAPI, "quests", "Quests");
            questAreaMarkerSet = getMarkerSet(dynmapAPI, "quest_areas", "Quest areas");
            if (questMarkerSet != null && questAreaMarkerSet != null) {
                questMarkerSet.setHideByDefault(false);
                questAreaMarkerSet.setHideByDefault(true);
                isEnabled = true;
                return;
            }
        }
        isEnabled = false;
    }

    @Override
    public UpdateRate getUpdateRate() {
        return UpdateRate.VERY_SLOW;
    }

    @Override
    public void updateMarkers() {
        if (questsAPI.isLoading()) {
            return;
        }

        //Remove old quest markers
        questMarkerSet.getMarkers().forEach(GenericMarker::deleteMarker);
        questAreaMarkerSet.getCircleMarkers().forEach(GenericMarker::deleteMarker);

        for (IQuest quest : questsAPI.getLoadedQuests()) {
            UUID npcStartId = quest.getNpcStart();
            if (npcStartId != null) {
                String markerDescription = "<b>Quest name:</b> " + quest.getName() + "<br><b>Quest description:</b> " + quest.getDescription();
                addNPCMarker(npcStartId, "Quest Start NPC: ", markerDescription, markerIcons.get(Icon.QUEST_GIVER), questMarkerSet);
            }
            for (IStage stage : quest.getStages()) {
                markNPCsInQuest(quest, stage, npcStartId);
                markKillLocations(quest, stage);
                markReachLocations(quest, stage);
                //TODO: Mark WorldGuard areas part of quests. Requires WorldGuard integration
                //TODO: See if there is anything to do against overlapping markers
            }
        }
    }

    /**
     * Marks any reach locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for reach locations</p>
     */
    private void markReachLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToReach(), stage.getRadiiToReachWithin(),
                "<b>Target location for:</b> " + quest.getName() + "<br><b>Description:</b> " +
                        quest.getDescription());
    }

    /**
     * Marks any kill locations found in the given stage
     *
     * @param quest <p>The quest the stage belongs to</p>
     * @param stage <p>The stage to search for kill locations</p>
     */
    private void markKillLocations(IQuest quest, IStage stage) {
        markLocations(stage.getLocationsToKillWithin(), stage.getRadiiToKillWithin(),
                "<b>Kill location for:</b> " + quest.getName() + "<br><b>Description:</b> " +
                        quest.getDescription());
    }

    /**
     * Marks the given locations on the dynamic map
     *
     * @param locations   <p>The locations to mark</p>
     * @param radii       <p>The radius of each location's circle</p>
     * @param description <p>The description for what the location means</p>
     */
    private void markLocations(List<Location> locations, List<Integer> radii, String description) {
        for (int i = 0; i < locations.size(); i++) {
            Location location = locations.get(i);
            int radius = radii.get(i);

            //Skip if location is invalid
            World world = location.getWorld();
            if (world == null) {
                continue;
            }

            CircleMarker circleMarker = questAreaMarkerSet.createCircleMarker(null, description, true,
                    world.getName(), location.getX(), location.getY(), location.getZ(), radius, radius, false);
            if (circleMarker == null) {
                DynmapCitizens.getInstance().getLogger().log(Level.WARNING, "Unable to create circle marker at " +
                        location + " with radius " + radius);
            } else {
                circleMarker.setFillStyle(0.3, 0x75AFD2);
                circleMarker.setLineStyle(1, 1.0, 0x36c90e);
            }
        }
    }

    /**
     * Marks all NPCs which has interactions as part of this stage
     *
     * @param quest      <p>The quest to mark NPCs for</p>
     * @param stage      <p>The quest stage to mark NPCs for</p>
     * @param npcStartId <p>The id of the quest's start NPC</p>
     */
    private void markNPCsInQuest(IQuest quest, IStage stage, UUID npcStartId) {
        String markerDescription = "Part of quest: " + quest.getName();
        //Mark all interaction targets
        for (UUID interactNpcId : stage.getNpcsToInteract()) {
            addNPCMarker(interactNpcId, "Quest Interact NPC: ", markerDescription,
                    markerIcons.get(Icon.QUEST_INTERACT), questMarkerSet);
        }
        //Mark all kill targets
        for (UUID killNPCId : stage.getNpcsToKill()) {
            addNPCMarker(killNPCId, "Quest Kill NPC: ", markerDescription, markerIcons.get(Icon.QUEST_KILL),
                    questMarkerSet);
        }
        //Mark all delivery targets
        for (UUID deliveryTargetNPCId : stage.getItemDeliveryTargets()) {
            if (!deliveryTargetNPCId.equals(npcStartId)) {
                addNPCMarker(deliveryTargetNPCId, "Quest Delivery NPC: ", markerDescription,
                        markerIcons.get(Icon.QUEST_DELIVER), questMarkerSet);
            }
        }
    }

}
